﻿using NUnit.Framework;
using Tile16_Decoupled;

namespace Test_Tile16_Decoupled
{
    [TestFixture]
    public class TestTileMover
    {
        private Board board;
        private TileMover tileMover;
        private Tile firstTile;
        private readonly Point firstPoint = new Point(0, 0);
        private Tile secondTile;
        private readonly Point secondPoint = new Point(1,1);

        [SetUp]
        public void Setup()
        {
            board = new Board(4, 4);
            firstTile = new Tile(1);
            secondTile = new Tile(2);
            board.AddTile(firstTile, firstPoint);
            tileMover = new TileMover(board);
            Assert.That(board.TileExists(firstPoint), Is.True);
        }

        [Test]
        public void MoveInTheCurrentLocation()
        {
            tileMover.Move(firstPoint, firstPoint);
            Assert.That(board.TileExists(firstPoint), Is.True);
        }

        [Test]
        public void MoveTileOneStepFromCornerInEmptyGrid()
        {
            tileMover.Move(firstPoint, new Point(1, 1));
            Assert.That(board.TileExists(firstPoint), Is.False);
            Assert.That(board.TileExists(new Point(1, 1)), Is.True);
        }

        [Test]
        public void MoveTileToInvalidAreasShouldThrow()
        {
            board.AddTile(firstTile, new Point(1, 1));
            Assert.Throws<InvalidBoardCoordinatesExcpetion>(() => tileMover.Move(firstPoint, new Point(-1, -1)));
            Assert.Throws<InvalidBoardCoordinatesExcpetion>(() => tileMover.Move(firstPoint, new Point(4, 4)));

        }

        [Test]
        public void MoveToExistingTileLocationWithDifferentValueNothingShouldHappen()
        {
            secondTile = new Tile(2);
            board.AddTile(secondTile, secondPoint);
            Assert.That(board.TileExists(secondPoint), Is.True);
            tileMover.Move(firstPoint, secondPoint);
            Assert.That(board.GetTile(firstPoint), Is.EqualTo(firstTile));
            Assert.That(board.GetTile(secondPoint), Is.EqualTo(secondTile));
        }

        [Test]
        public void MoveToExistingTileLocationWithSameValueShouldCombine()
        {
            board.AddTile(firstTile, secondPoint);
            tileMover.Move(firstPoint, secondPoint);
            Assert.That(board.TileNotExists(firstPoint), Is.True);
            Assert.That(board.GetTile(secondPoint), Is.EqualTo(new Tile(2)));
        }

        [Test]
        public void MovingATileWithoutAValidPathNothingShouldHappen()
        {
            var mockPathFinder = new MockPathFinder();
            tileMover.SetPathFinder(mockPathFinder);
            tileMover.Move(firstPoint, secondPoint);
            Assert.That(board.GetTile(firstPoint), Is.EqualTo(firstTile));
            Assert.That(board.TileNotExists(secondPoint), Is.True);
            Assert.That(mockPathFinder.PathExistsIsSet, Is.True);
        }
    }

    public class MockPathFinder : BoardPathFinder
    {
        public bool PathExistsIsSet = false;

        public bool PathExists(Point form, Point to)
        {
            PathExistsIsSet = true;
            return false;
        }
    }
}