﻿using System.Linq;
using NUnit.Framework;
using Tile16_Decoupled;

namespace Test_Tile16_Decoupled
{
    [TestFixture]
    public class TestTileSpawner
    {
        private Board board;
        private TileSpawner spawner;

        public class NonRandomBoard : Board
        {
            private readonly int width;
            private readonly int height;
            private int x = 0;
            private int y = 0;
            public NonRandomBoard(int width, int height) : base(width, height)
            {
                this.width = width;
                this.height = height;
            }

            protected override Point GenerateRandomUnoccupiedPoint()
            {
                if (x == width)
                {
                    x = 0;
                    y++;
                }
                return new Point(x++, y);
            }
        }

        public class NonRandomTileSpawner : TileSpawner
        {
            private int tileValue = 2;

            public NonRandomTileSpawner(Board board) : base(board)
            {
            }

            protected override Tile GenerateRandomBoundedTile()
            {
                return new Tile(tileValue++);
            }
        }

        [SetUp]
        public void Setup()
        {
            board = new NonRandomBoard(4, 4);
            spawner = new TileSpawner(board);
        }

        [Test]
        public void SpawnRandomTilesOnBoard()
        {
            spawner.Spawn(0);
            Assert.That(board.Empty(), Is.True);
            spawner.Spawn(1);
            Assert.That(board.Empty(), Is.False);
            spawner.Spawn(1);
            Assert.That(board.TileCount(), Is.EqualTo(2));
            spawner.Spawn(2);
            Assert.That(board.TileCount(), Is.EqualTo(4));
            spawner.Spawn(12);
            Assert.That(board.TileCount(), Is.EqualTo(16));
        }
        
        [Test]
        public void SpawnRandomTilesWithNotEnoughUnoccupiedSpacesShouldBeOk()
        {
            spawner.Spawn(14);
            Assert.That(board.TileCount(), Is.EqualTo(14));
            spawner.Spawn(3);
            Assert.That(board.TileCount(), Is.EqualTo(16));
        }

        [Test]
        public void SpawnedValuesAreBounded()
        {   
            AddTwoSameTilesToTheEndOfTheBoard();
            SpawnClampedTile(1);
            Assert.That(board.GetTile(new Point(0, 0)), Is.EqualTo(new Tile(1)));
            SpawnClampedTile(2);
            Assert.That(board.GetTile(new Point(1, 0)), Is.EqualTo(new Tile(2)));
            SpawnClampedTile(5);
            Assert.That(board.GetTile(new Point(2, 0)), Is.EqualTo(new Tile(5)));
        }

        private void AddTwoSameTilesToTheEndOfTheBoard()
        {
            board.AddTile(new Tile(1), new Point(2, 3));
            board.AddTile(new Tile(1), new Point(3, 3));
        }

        private void SpawnClampedTile(int clampValue)
        {
            spawner.SetMaxSpawnTileValue(clampValue);
            spawner.SetMinSpawnTileValue(clampValue);
            spawner.Spawn(1);
        }

        [Test]
        public void SpawnedTilesMustHaveOneThatFormAPairWithBoardOrItself()
        {
            Point[] point = new Point[16];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                    point[i * 4 + j] = new Point(j, i);

            spawner = new NonRandomTileSpawner(board);
            spawner.SetMaxSpawnTileValue(16);
            spawner.SetMinSpawnTileValue(0);
            spawner.Spawn(2);   //empty board shoudl spawn two same ones
            Assert.That(board.GetTile(point[0]), Is.EqualTo(board.GetTile(point[1])));
            spawner.Spawn(1); //force spawning a different one must work
            Assert.That(board.GetTile(point[2]), Is.Not.EqualTo(board.GetTile(point[1])));
            board.PopTile(point[0]);
            spawner.Spawn(1); // two different tiles spawn must contain
            AssertTileValueHasPair(point[3]);
        }


        private void AssertTileValueHasPair(Point point)
        {
            var tileList = board.GetTileList();
            Assert.That(tileList.Count(i => i.Equals(board.GetTile(point))), Is.GreaterThan(1));
        }
    }

    
}