﻿using NUnit.Framework;
using Tile16_Decoupled;
using Tile16_Unity_Testable;
using UnityEngine;

namespace Test_Tile16_Unity_Testable
{
    
    public class TestGameController
    {
        private GameController gameController;
        private MockGameControllerMediator mockMediator;

        [SetUp]
        public void Setup()
        {
            TileController.OnScreenTileList.Clear();
            TileController.TileControllerDictionary.Clear();
            gameController = new GameController();
            mockMediator = new MockGameControllerMediator(gameController);
            gameController.SetMediator(mockMediator);
        }

        private class TestNewGameInitialization : TestGameController
        {
            [Test]
            public void NewGameShouldRemoveAllTiles()
            {
                var tile1 = new TileController();
                var mockTileInputControllerMediator = new MockTileInputControllerMediator(tile1);
                tile1.SetMediator(mockTileInputControllerMediator);
                var tile2 = new TileController();
                tile2.SetMediator(new MockTileInputControllerMediator(tile2));
                Assert.That(TileController.OnScreenTileList.Count, Is.EqualTo(2));
                gameController.NewGame();
                Assert.That(TileController.OnScreenTileList.Count, Is.EqualTo(0));
            }
        }

        [TestFixture]
        class TestNewGames : TestGameController
        {
            private MockWaveController mockWaveController;
            private MockWinConditions mockWinConditions;
            [SetUp]
            public void Setup()
            {
                mockWaveController = new MockWaveController();
                mockWinConditions = new MockWinConditions();
                gameController.SetWaveController(mockWaveController);
                gameController.SetWinCondition(mockWinConditions);
                gameController.NewGame();
            }

            [Test]
            public void OnGameOverDisableTileController()
            {
                Assert.That(TileController.Enabled, Is.True);
                mockWinConditions.IsGameWonReturn = true;
                gameController.Update();
                Assert.That(TileController.Enabled, Is.False);
                var tile1 = new TileController();
                var mockTileInputControllerMediator = new MockTileInputControllerMediator(tile1);
                tile1.SetMediator(mockTileInputControllerMediator);
                tile1.PickUp();
                Assert.That(tile1.IsPickedUp, Is.False);
                Assert.That(TileController.Enabled, Is.False);
            }

            [Test]
            public void NewGameCreatesBoardAndResetsTime()
            {
                Assert.That(TileController.TileControllerDictionary.Count, Is.EqualTo(0));
                Assert.That(gameController.GameBoard.Empty(), Is.True);
                Assert.That(GameController.Time, Is.EqualTo(0));
                //Assert.That(mockMediator.UpdateBoardIsCalled, Is.True);
            }


            [Test]
            public void OnFrameUpdateDecrementTimerWhenBoardHasDuplicate()
            {
                GameController.Time = 1.0f;
                gameController.GameBoard.AddTile(new Tile(), new Point());
                gameController.GameBoard.AddTile(new Tile(), new Point(1,0));
                gameController.Update();
                Assert.That(GameController.Time, Is.EqualTo(0.9f).Within(0.0001));
            }

            [Test]
            public void NextWaveShouldPlaySoundAndAnimation()
            {
                mockWaveController.NextWaveReturn = true;
                gameController.Update();
                Assert.That(mockMediator.NextWaveSfxPlayed, Is.True);
            }

            [Test]
            public void NextWaveShouldUseWaveControllerAndUpdateBoard()
            {
                mockWaveController.NextWaveReturn = true;
                gameController.Update();
                Assert.That(mockWaveController.NextWaveIsCalled, Is.True);
                Assert.That(mockMediator.UpdateBoardIsCalled, Is.True);
            }

            [Test]
            public void ProceedToNextWaveIfNoDuplicate()
            {
                GameController.Time = 0.5f;
                gameController.Update();
                Assert.That(mockWaveController.NextWaveIsCalled, Is.True);
            }

            [Test]
            public void OnGameWinPlaySfxAndAnimation()
            {
                mockWinConditions.IsGameWonReturn = true;
                Assert.That(mockMediator.WinAnimationEnabled, Is.False);
                gameController.Update();
                Assert.That(mockMediator.WinSfxPlayed, Is.True);
                Assert.That(mockMediator.WinAnimationEnabled, Is.True);

            }

            [Test]
            public void CannotProgressToNextWaveShouldLose()
            {
                mockWinConditions.GameOverButDidntWinReturn = true;
                gameController.Update();
                Assert.That(mockWinConditions.GameOverButDidntWinIsCalled, Is.True);
            }

            [Test]
            public void GameOverShouldPlayLoseAnimationAndSound()
            {
                mockWinConditions.GameOverButDidntWinReturn = true;
                gameController.Update();
                Assert.That(mockMediator.LoseSfxPlayed, Is.True);
                Assert.That(mockMediator.LoseAnimationEnabled, Is.True);
            }

            [Test]
            public void OnNextWavePickedUpTileShouldRestorePosition()
            {   
                GlobalTileMover.tileMover = new TileMover(gameController.GameBoard);
                var tileInputController = new TileController();
                TileController.TileControllerDictionary.Clear();
                TileController.OnScreenTileList.Clear();
                var mockTileMediator = new MockTileInputControllerMediator(tileInputController);
                tileInputController.SetMediator(mockTileMediator);
                tileInputController.InitTile(new Tile(), new Point(0, 0));

                Vector2 lastPosition = mockTileMediator.Position;
                tileInputController.PickUp();
                mockTileMediator.Position = new Vector2(200f,200f);
                gameController.Update();
                Assert.That(tileInputController.IsPickedUp, Is.False);
                Assert.That(mockTileMediator.UpdateBoardIsCalled, Is.False);
                Assert.That(mockTileMediator.LastSetPosition, Is.EqualTo(lastPosition));
            }

            [Test]
            public void NewGameShouldDisableLoseOrWinAnimation()
            {
                Assert.That(mockMediator.DisableWinAnimationIsCalled, Is.True);
                Assert.That(mockMediator.DisableLoseAnimationIsCalled, Is.True);
            }

            [Test]
            public void GameOverShouldStopUpdate()
            {
                mockWinConditions.GameOverButDidntWinReturn = true;
                gameController.Update();
                mockWinConditions.GameOver = true;
                gameController.Update();
                Assert.That(mockWinConditions.IsGameWonCalledOnce, Is.True);
            }
        }
       


    }

    internal class MockWinConditions : IWinConditions
    {
        public bool IsGameWonReturn;
        public bool GameOverButDidntWinIsCalled;
        public bool GameOverButDidntWinReturn;
        public bool IsGameWonCalledOnce ;
        private int gameWonCalledCounter = 0;
        public bool GameOver { get; set; }
        public bool GameWon { get; }
        public int WinningTileValue { get; set; }


        public bool GameOverButDidntWin()
        {
            GameOverButDidntWinIsCalled = true;
            return GameOverButDidntWinReturn;
        }

        public bool IsGameWon()
        {
            gameWonCalledCounter++;
            IsGameWonCalledOnce = gameWonCalledCounter == 1 ;
            return IsGameWonReturn;
        }

        public bool NotWinningButNotGameOver()
        {
            throw new System.NotImplementedException();
        }

        public void Reset()
        {
            throw new System.NotImplementedException();
        }

        public void SetBoard(Board board)
        {
            throw new System.NotImplementedException();
        }

        public void UpdateStatusByBiggestTileValue()
        {
            throw new System.NotImplementedException();
        }
    }

    internal class MockWaveController : IWaveController
    {
        public bool NextWaveIsCalled;
        public int MaxTileSpawnCount { get; set; }
        public bool NextWaveReturn;
        public void InitializeNewGame()
        {
            throw new System.NotImplementedException();
        }

        public bool NextWave()
        {
            NextWaveIsCalled = true;
            return NextWaveReturn;
        }

        public void SetBoard(Board board)
        {
            throw new System.NotImplementedException();
        }

        public void SetWinConditions(WinConditions winConditions)
        {
            throw new System.NotImplementedException();
        }
    }

    internal class MockGameControllerMediator : GameControllerMediator
    {
        private GameController gameController;
        public bool InitializeIsCalled;
        public bool UpdateBoardIsCalled;
        public bool NextWaveSfxPlayed;
        public bool WinSfxPlayed;
        public bool WinAnimationEnabled;
        public bool LoseSfxPlayed;
        public bool LoseAnimationEnabled;
        public bool DisableWinAnimationIsCalled;
        public bool DisableLoseAnimationIsCalled;

        public MockGameControllerMediator(GameController gameController)
        {
            this.gameController = gameController;
        }

        public void Initialize()
        {
            InitializeIsCalled = true;
        }

        public void UpdateBoard()
        {
            UpdateBoardIsCalled = true;
        }

        public void PlayLoseSfx()
        {
            LoseSfxPlayed = true;
        }

        public void EnableWinAnimation()
        {
            WinAnimationEnabled = true;
        }

        public void EnableLoseAnimation()
        {
            LoseAnimationEnabled = true;
        }

        public void DisableWinAnimation()
        {
            DisableWinAnimationIsCalled = true;
        }

        public void DisableLoseAnimation()
        {
            DisableLoseAnimationIsCalled = true;
        }

        public float GetDeltaTime()
        {
            return 0.1f;
        }

        public void PlayNextWaveSfx()
        {
            NextWaveSfxPlayed = true;
        }

        public void PlayWinSfx()
        {
            WinSfxPlayed = true;
        }
    }
}