﻿using NUnit.Framework;
using Tile16_Decoupled;
using Tile16_Unity_Testable;
using UnityEngine;

namespace Test_Tile16_Unity_Testable
{
    
    public class TestTileController
    {
        private TileController tileInputController;
        private MockTileInputControllerMediator mockMediator;
        private TileController target;
        private MockTileInputControllerMediator mockTargetMediator;

        [SetUp]
        public void SetupOneTileController()
        {
            var board = new Board(4, 4);
            GlobalTileMover.tileMover = new TileMover(board);
            tileInputController = new TileController();
            TileController.TileControllerDictionary.Clear();
            TileController.OnScreenTileList.Clear();
            mockMediator = new MockTileInputControllerMediator(tileInputController);
            tileInputController.SetMediator(mockMediator);
            tileInputController.InitTile(new Tile(), new Point(0,0));
        }

        [TestFixture]
        class TestWithMockInputControllerMediator : TestTileController
        {
            [Test]
            public void OnCreateShouldPlaySpawnAnimation()
            {
                Assert.That(mockMediator.SpawnAnimationPlayed, Is.True);
            }

            [Test]
            public void OnCreateShouldSetTileToKinematic()
            {
                Assert.That(mockMediator.IsKinematicSet, Is.True);
            }


            [Test]
            public void PickUpTileShouldPlaySoundAnimationAndSetPosition()
            {
                //mockMediator.MouseDown();
                tileInputController.PickUp();
                Assert.That(mockMediator.PickUpSfxPlayed, Is.True);
                Assert.That(mockMediator.PickUpAnimationPlayed, Is.True);
                Assert.That(tileInputController.IsPickedUp, Is.True);
                Assert.That(tileInputController.PickedUpPosition, Is.EqualTo(new Vector2(50f, 50f)));
            }

            [Test]
            public void FollowMouseShouldChangeTileVelocity()
            {
                Assert.That(tileInputController.Velocity, Is.EqualTo(Vector2.zero));
                tileInputController.PickUp();
                tileInputController.FollowMouse(new Vector2(100f, 100f));
                Assert.That(tileInputController.Velocity, Is.Not.EqualTo(Vector2.zero));
            }

            [Test]
            public void DropTileShouldPlaySoundAnimation()
            {
                tileInputController.PickUp();
                tileInputController.Drop();
                Assert.That(mockMediator.DropSfxPlayed, Is.True);
                Assert.That(mockMediator.DropAnimationPlayed, Is.True);
                Assert.That(tileInputController.IsPickedUp, Is.False);
            }

            [Test]
            public void DropTileShouldSnapToGridAndMoveTileOnBardAndZeroOutVelocity()
            {
                tileInputController.PickUp();
                tileInputController.Drop();
                Assert.That(tileInputController.IsPickedUp, Is.False);
                Assert.That(mockMediator.GetVelocity(), Is.EqualTo(Vector2.zero));
                Assert.That(mockMediator.LastSetPosition, Is.EqualTo(new Vector2(52f, 52f)));
                Assert.That(tileInputController.Velocity, Is.EqualTo(Vector2.zero));
            }

            [Test]
            public void OnDropShouldSnapToNearestGrid()
            {
                tileInputController.PickUp();
                mockMediator.Position = new Vector2(99.0f, 99.0f);
                tileInputController.Drop();
                Assert.That(mockMediator.LastSetPosition, Is.EqualTo(new Vector2(52f, 52f)));
            }

            [Test]
            public void OnDropShouldMoveCurrentPoint()
            {
                tileInputController.PickUp();
                mockMediator.Position = new Vector2(110.0f, 10.0f);
                tileInputController.Drop();
                Assert.That(mockMediator.LastSetPosition, Is.EqualTo(new Vector2(152f, 52f)));
                
            }

            [Test]
            public void OnDropShouldSetKinematicToTrue()
            {
                tileInputController.PickUp();
                tileInputController.Drop();
                Assert.That(mockMediator.IsKinematicSet, Is.True);
            }

            

            [Test]
            public void TileShouldNotMoveIfNotPickedUp()
            {
                tileInputController.FollowMouse(new Vector2(200f,200f));
                Assert.That(tileInputController.Velocity, Is.EqualTo(Vector2.zero));
            }
        }
       
        [TestFixture]
        public class DestroyedTileTests : TestTileController
        {
            [SetUp]
            public void SetupTileAlreadyDestroyed()
            {
                tileInputController.Destroyed = true;
            }

            [Test]
            public void IfDestroyedThenCannotBePickedUp()
            {
                tileInputController.PickUp();
                Assert.That(tileInputController.IsPickedUp, Is.False);
            }

            [Test]
            public void IfDestroyedThenCannotMove()
            {
                tileInputController.FollowMouse(new Vector2(200f, 200f));
                Assert.That(tileInputController.Velocity, Is.EqualTo(Vector2.zero));
            }

            [Test]
            public void IfDestroyedThenCannotDrop()
            {
                tileInputController.IsPickedUp = true;
                tileInputController.Drop();
                Assert.That(tileInputController.IsPickedUp, Is.True);
            }

        }

        class MultiTileTests : TestTileController
        {
            [SetUp]
            public void SetupTargetTile()
            {
                target = new TileController();
                mockTargetMediator = new MockTileInputControllerMediator(target);
                target.SetMediator(mockTargetMediator);
                target.InitTile(new Tile(), new Point(1, 1));
            }

            [Test]
            public void CombineTilesShouldWhat()
            {
                Assert.That(TileController.TileControllerDictionary[new Tile()].Count, Is.EqualTo(2));
                tileInputController.Combine(target);
                Assert.That(mockMediator.CombineSfxPlayed, Is.True);
                Assert.That(mockMediator.TargetRendererIsInactive, Is.True);
                Assert.That(mockMediator.DestroyCollisionObjectsIsCalled, Is.True);
                Assert.That(mockMediator.UpdateBoardIsCalled, Is.True);
                Assert.That(tileInputController.Destroyed, Is.True);
                Assert.That(target.Destroyed, Is.True);
                Assert.That(TileController.TileControllerDictionary.ContainsKey(new Tile()), Is.False);
            }

            [Test]
            public void OnPickupShrinkOtherTilesOfSameValueColliderSmaller()
            {
                tileInputController.PickUp();
                Assert.That(mockMediator.IsKinematicUnset, Is.True);
                Assert.That(mockMediator.LastSetColliderSize, Is.EqualTo(new Vector2(1f, 1f)));
                Assert.That(mockTargetMediator.LastSetColliderSize, Is.EqualTo(new Vector2(0.7f, 0.7f)));
            }

            [Test]
            public void IfDestroyedThenCannotCombine()
            {
                target.Destroyed = true;
                tileInputController.Combine(target);
                Assert.That(tileInputController.Destroyed, Is.False);
            }

            [Test]
            public void DontCombineIfTileValuesAreDifferent()
            {
                target.InitTile(new Tile(2), new Point(2,2));
                tileInputController.Combine(target);
                Assert.That(tileInputController.Destroyed, Is.False);
                Assert.That(target.Destroyed, Is.False);
            }

            [Test]
            public void OnDropShouldRestoreColliderForSameValueTiles()
            {
                tileInputController.PickUp();
                tileInputController.Drop();
                Assert.That(mockMediator.LastSetColliderSize, Is.EqualTo(new Vector2(1f, 1f)));
                Assert.That(mockTargetMediator.LastSetColliderSize, Is.EqualTo(new Vector2(1f, 1f)));
            }

        }
    }

    public class MockTileInputControllerMediator : TileControllerMediator
    {
        private TileController tileInputController;
        public bool PickUpSfxPlayed;
        public bool PickUpAnimationPlayed;
        public Vector2 Position = new Vector2(50f, 50f);
        
        private Vector2 velocity = Vector2.zero;
        public bool DropSfxPlayed;
        public bool DropAnimationPlayed;
        public bool DestroyCollisionObjectsIsCalled;
        public bool UpdateBoardIsCalled;
        public bool CombineSfxPlayed;
        public Vector2 LastSetColliderSize;
        public Vector2 LastSetPosition;
        public bool IsKinematicUnset;
        public bool IsKinematicSet;
        public bool SpawnAnimationPlayed;
        public bool TargetRendererIsInactive;

        public MockTileInputControllerMediator(TileController tileInputController)
        {
            this.tileInputController = tileInputController;
        }

        public Vector2 GetPosition()
        {
            return Position;
        }

        public Vector2 GetVelocity()
        {
            return velocity;
        }

        public void SetVelocity(Vector2 velocity)
        {
            this.velocity = velocity;
        }

        public void SetPosition(Vector2 vector2)
        {
            LastSetPosition = vector2;
        }

        public float GetScaleX()
        {
            return 1.0f;
        }

        public float GetBoundSizeX()
        {
            return 100.0f;
        }

        public void SetRendererInactive()
        {
            TargetRendererIsInactive = true;
        }

        public void MouseDown()
        {
            tileInputController.PickUp();
        }

        public void MoveMouse(Vector2 to)
        {
            tileInputController.FollowMouse(to);
        }

        public void PlayPickUpAnimation()
        {
            PickUpSfxPlayed = true;
        }

        public void PlayPickUpSfx()
        {
            PickUpAnimationPlayed = true;
        }

        public void PlayDropSfx()
        {
            DropSfxPlayed = true;
        }

        public void PlayDropAnimation()
        {
            DropAnimationPlayed = true;
        }

        public void PlayCombineSfx()
        {
            CombineSfxPlayed = true;
        }

        public void PlaySpawnAnimation()
        {
            SpawnAnimationPlayed = true;
        }

        public void KinematicUnset()
        {
            IsKinematicUnset = true;
        }

        public void UpdateBoard()
        {
            UpdateBoardIsCalled = true;
        }

        public void DestroyCollisionObjects()
        {
            DestroyCollisionObjectsIsCalled = true;
        }

        public void DestroySelf()
        {
            
        }

        public void KinematicSet()
        {
            IsKinematicSet = true;
        }

        public void SetColliderSize(Vector2 vector2)
        {
            LastSetColliderSize = vector2;
        }
    }
}