﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Tile16_Decoupled
{
    public class InvalidBoardDimensionsException : Exception { }
    public class InvalidBoardCoordinatesExcpetion : Exception { }
    public class BoardLocationOccupiedException : Exception { }
    public class BoardLocationNotOccupiedException : Exception { }

    public class Board
    {
        private readonly int width;
        private readonly int height;
        private readonly Dictionary<Point, Tile> occupiedGrid;
        public Dictionary<Point, Tile> newlyAddedTiles;
        private Dictionary<Point, object> unoccupiedGrid;
        private readonly Dictionary<int, int> tileValueCount;
        private int biggestTileValue = 1;

        public Board(int width, int height)
        {
            AssertBoardDimensions(width, height);
            this.width = width;
            this.height = height;
            occupiedGrid = new Dictionary<Point, Tile>();
            newlyAddedTiles = new Dictionary<Point, Tile>();
            InitializeUnoccupiedGrid();
            tileValueCount = new Dictionary<int, int>();
        }

        private void InitializeUnoccupiedGrid()
        {
            unoccupiedGrid = new Dictionary<Point,object>();
            for (int i = 0; i < width; i++)
                for (int j = 0; j < height; j++)
                    unoccupiedGrid.Add(new Point(i, j), null);
        }

        private static void AssertBoardDimensions(int width, int height)
        {
            if (width <= 0 || height <= 0)
                throw new InvalidBoardDimensionsException();
        }

        public int GetWidth()
        {
            return width;
        }

        public int GetHeight()
        {
            return height;
        }

        public void AddTile(Tile tile, Point point)
        {
            AssertCoordinatesNotOutOfRange(point);
            AssertCoordindateNotOccupied(point);
            occupiedGrid[point] = tile;
            newlyAddedTiles[point] = tile;
            unoccupiedGrid.Remove(point);
            AddTileValueCount(tile);
            UpdateBiggestTileValue(tile);
        }

        private void UpdateBiggestTileValue(Tile tile)
        {
            if (tile.GetValue() > biggestTileValue)
                biggestTileValue = tile.GetValue();
        }

        private void AddTileValueCount(Tile tile)
        {
            int tileValue = 0;
            if (tileValueCount.TryGetValue(tile.GetValue(), out tileValue))
                tileValueCount[tile.GetValue()] = tileValue + 1;
            else
                tileValueCount[tile.GetValue()] = 1;
        }

        private void AssertCoordindateNotOccupied(Point point)
        {
            if (TileExists(point))
                throw new BoardLocationOccupiedException();
        }

        private void AssertCoordinatesNotOutOfRange(Point point)
        {
            if (point.X < 0 || point.Y < 0 || point.X >= width || point.Y >= height)
                throw new InvalidBoardCoordinatesExcpetion();
        }

        public bool Empty()
        {   
            return occupiedGrid.Count == 0;
        }

        public int TileCount()
        {   
            return occupiedGrid.Count;
        }

        public bool TileExists(Point point)
        {
            AssertCoordinatesNotOutOfRange(point);
            return occupiedGrid.ContainsKey(point);
        }

        public bool TileNotExists(Point point)
        {
            AssertCoordinatesNotOutOfRange(point);
            return unoccupiedGrid.ContainsKey(point);
        }

        public void AddTileToRandomLocation(Tile tile)
        {
            AddTile(tile, GenerateRandomUnoccupiedPoint());
        }

        protected virtual Point GenerateRandomUnoccupiedPoint()
        {
            var rand = new Random();
            var toList =  new List<Point>(unoccupiedGrid.Keys);
            var randPoint = toList[rand.Next(unoccupiedGrid.Count - 1)];
            return randPoint;
        }

        public int NumberOfUnoccupiedSpace()
        {
            return unoccupiedGrid.Count;
        }

        public Tile GetTile(Point point)
        {
            AssertCoordinatesNotOutOfRange(point);
            AssertCoordindateOccupied(point);
            return occupiedGrid[point];
        }

        private void AssertCoordindateOccupied(Point point)
        {
            if (TileNotExists(point))
                throw new BoardLocationNotOccupiedException();
        }

        public Tile PopTile(Point point)
        {
            var tile = GetTile(point);
            DecrementTileValueCount(tile);
            occupiedGrid.Remove(point);
            unoccupiedGrid.Add(point, null);
            return tile;
        }

        private void DecrementTileValueCount(Tile tile)
        {
            tileValueCount[tile.GetValue()]--;
            if (tileValueCount[tile.GetValue()] == 0)
                tileValueCount.Remove(tile.GetValue());
        }

        public IList<Tile> GetTileList()
        {
            var toList = new List<Tile>(occupiedGrid.Values);
            return toList.AsReadOnly();
        }

        public bool HasDuplicate()
        {
            return tileValueCount.Values.Any(count => count > 1);
        }

        public int GetBiggestTileValue()
        {
            return biggestTileValue;
        }

        public bool IsFull()
        {
            return TileCount() == width*height;
        }

        public Dictionary<Point, Tile> PopNewlyCreatedTiles()
        {
            var temp = newlyAddedTiles;
            newlyAddedTiles = new Dictionary<Point, Tile>();
            return temp;
        }

        public void MoveTile(Point from, Point to)
        {
            if (from.Equals(to)) return;
            if (TileNotExists(from)) return;
            var tile = occupiedGrid[from];
            occupiedGrid.Remove(from);
            occupiedGrid[to] = tile;
            unoccupiedGrid[from] = null;
            unoccupiedGrid.Remove(to);
        }
    }
}