﻿using System.Collections.Generic;
using Tile16_Decoupled;
using UnityEngine;

namespace Tile16_Unity_Testable
{
    public class TileController
    {
        private TileControllerMediator mediator;// = new NullTICMediator();

        public static List<TileController> OnScreenTileList = new List<TileController>();
        public static Dictionary<Tile, List<TileController>> TileControllerDictionary =
            new Dictionary<Tile, List<TileController>>();

        public bool IsPickedUp;
        public Vector2 PickedUpPosition;
        public Vector2 Velocity;
        public float MoveSpeed = 20.0f;
        public float MaxSpeed = 30.0f;
        public bool Destroyed;

        private Tile tile;
        private Point pos;
        private TileSnappingGrid snappingGrid;
        private static TileController pickedUpTile;
        public static bool Enabled;

        public TileController()
        {
            OnScreenTileList.Add(this);
        }

        public void InitTile(Tile tile, Point pos)
        {
            this.tile = tile;
            this.pos = pos;

            snappingGrid = new TileSnappingGrid();
            //spriteRenderer.bounds.size.x / transform.localScale.x);
            snappingGrid.SetTileSideLength(mediator.GetBoundSizeX()/mediator.GetScaleX());
            snappingGrid.InitializeGridWithDimensions(4, 4);
            AddTileControllerToDictionary();
            mediator.PlaySpawnAnimation();
            mediator.KinematicSet();
        }

        public void AddTileControllerToDictionary()
        {
            List<TileController> tileControllers;
            if (!TileControllerDictionary.TryGetValue(tile, out tileControllers))
            {
                tileControllers = new List<TileController>();
                TileControllerDictionary.Add(tile, tileControllers);
            }
            tileControllers.Add(this);
        }

        public void SetMediator(TileControllerMediator mediator)
        {
            this.mediator = mediator;
            Velocity = mediator.GetVelocity();
        }

        public void PickUp()
        {
            if (Destroyed || !Enabled) return;
            IsPickedUp = true;
            pickedUpTile = this;
            PickedUpPosition = mediator.GetPosition();
            mediator.PlayPickUpAnimation();
            mediator.PlayPickUpSfx();
            PrepareTilesOfSameValue();
        }

        private void PrepareTilesOfSameValue()
        {
            mediator.KinematicUnset();
            foreach (var tileInputController in TileControllerDictionary[tile])
                tileInputController.mediator.SetColliderSize(new Vector2(0.7f, 0.7f));
            mediator.SetColliderSize(new Vector2(1.0f, 1.0f));
        }

        public void FollowMouse(Vector2 to)
        {
            if (!IsPickedUp || Destroyed) return;
            Velocity = (to - mediator.GetPosition())*MoveSpeed;
            if (Velocity.magnitude > MaxSpeed)
                Velocity *= MaxSpeed/Velocity.magnitude;
            mediator.SetVelocity(Velocity);
        }
        
        public void Drop()
        {
            if (!IsPickedUp || Destroyed) return;
            IsPickedUp = false;
            pickedUpTile = null;
            mediator.PlayDropSfx();
            mediator.PlayDropAnimation();
            mediator.KinematicSet();

            SnapToPosition();
            MoveBoardPoints();

            RevertTilesOfSameValue();

        }

        private void SnapToPosition()
        {
            mediator.SetPosition(
                snappingGrid.GetSnappingPosition(mediator.GetPosition() - GameConstants.gridLowerLeftCorner)
                                 + GameConstants.gridLowerLeftCorner);
        }

        public void Combine(TileController target)
        {
            TileController from = IsPickedUp ? this : target;
            TileController to = IsPickedUp ? target : this;

            if (from.Destroyed || to.Destroyed) return;
            if (!tile.Equals(target.tile)) return;

            from.Destroyed = true;
            to.Destroyed = true;

            mediator.SetRendererInactive();
            mediator.PlayCombineSfx();
            GlobalTileMover.tileMover.Move(from.pos, to.pos);
            mediator.UpdateBoard();
            mediator.DestroyCollisionObjects();

            RevertTilesOfSameValue();

            RemoveFromTileLists(from);
            RemoveFromTileLists(to);
        }

        private void MoveBoardPoints()
        {
            var newPoint = snappingGrid.GetSnappingPoint(mediator.GetPosition() - GameConstants.gridLowerLeftCorner);
            GlobalTileMover.tileMover.Move(pos, newPoint);
            pos = newPoint;
        }

        private void RemoveFromTileLists(TileController tileInputController)
        {
            OnScreenTileList.Remove(tileInputController);
            var dicList = TileControllerDictionary[tileInputController.tile];
            dicList.Remove(tileInputController);
            if (dicList.Count == 0)
                TileControllerDictionary.Remove(tileInputController.tile);
        }

        private void RevertTilesOfSameValue()
        {
            List<TileController> tileList;
            TileControllerDictionary.TryGetValue(tile, out tileList);
            if (tileList == null) return;
            foreach (var tileController in TileControllerDictionary[tile])
                tileController.mediator.SetColliderSize(new Vector2(1.0f, 1.0f));
        }

        public static void RestoreTileIfPickedUp()
        {
            if (pickedUpTile != null)
                pickedUpTile.DropWithOutMoving();
        }

        private void DropWithOutMoving()
        {
            if (!IsPickedUp || Destroyed) return;
            IsPickedUp = false;
            pickedUpTile = null;
            mediator.PlayDropSfx();
            mediator.PlayDropAnimation();
            mediator.KinematicSet();
            mediator.SetPosition(PickedUpPosition);
            RevertTilesOfSameValue();
        }

        public static void ClearAllTiles()
        {   
            foreach (var tileController in OnScreenTileList)
                tileController.mediator.DestroySelf();
            OnScreenTileList.Clear();
            TileControllerDictionary.Clear();
        
        }
    }

    public class GlobalTileMover
    {
        public static TileMover tileMover;
    }

    /*
    internal class NullTICMediator : TileInputControllerMediator
    {
    }
    */
}